"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Query = void 0;
const html_1 = require("./html");
const parser_1 = require("./parser");
const types_1 = require("./types");
const utils_1 = require("./utils");
class QueryInstance {
    constructor(document, matched, virtualDoms, previous) {
        this.getUntil = (selector, namespaces) => {
            return (typeof selector === "string"
                ? (() => {
                    const query = (0, parser_1.tryQueryParser)(selector);
                    if (query) {
                        return query.match(this.document.descendants(), this.document.descendants(), namespaces);
                    }
                    return [];
                })()
                : selector.matched).reduce((mapped, element) => {
                mapped[element.identifier] = true;
                return mapped;
            }, {});
        };
        this.addElementsToMatcher = (manipulator, firstArg, ...content) => {
            this.matched.forEach((element, index) => {
                const textFirstArg = typeof firstArg === "function"
                    ? firstArg(index, (0, utils_1.parserItemToString)(element))
                    : firstArg;
                const allContent = [textFirstArg, ...content];
                allContent.forEach((content, index) => {
                    manipulator(content, element, index);
                });
            });
            this.document.cache.descendants.invalid = true;
            return this;
        };
        this.getLowestChild = (element) => {
            if (element.children().length === 0) {
                return element;
            }
            return this.getLowestChild(element.children()[0]);
        };
        this.getParentAssigner = (resolved, namespaces) => {
            let parent = undefined;
            let appendTo = undefined;
            if (typeof resolved === "string") {
                const query = (0, parser_1.tryQueryParser)(resolved);
                if (query) {
                    parent = query.match(this.document.descendants(), this.document.descendants(), namespaces)[0];
                    appendTo = parent;
                }
                else {
                    const html = (0, parser_1.tryHtmlParser)(resolved, this.compress);
                    if (html) {
                        parent = html.descendants()[0];
                        appendTo = this.getLowestChild(parent);
                    }
                }
            }
            else {
                parent = resolved.matched[0]?.clone();
                appendTo = parent;
            }
            return [parent, appendTo];
        };
        this.filterBySelector = (selector, elements, namespaces) => {
            if (selector === undefined) {
                return createQuery(this.document, elements, this.virtualDoms, this);
            }
            if (typeof selector === "string") {
                const query = (0, parser_1.tryQueryParser)(selector);
                if (query) {
                    const matched = query.match(elements, this.document.descendants(), namespaces);
                    return createQuery(this.document, matched, this.virtualDoms, this);
                }
            }
            if (selector instanceof QueryInstance) {
                const map = selector.matched.reduce((map, element) => {
                    map[element.identifier] = true;
                    return map;
                }, {});
                const matched = elements.filter((n) => map[n.identifier]);
                return createQuery(this.document, matched, this.virtualDoms, this);
            }
            return this;
        };
        this.find = (queryInput, namespaces = {}) => {
            const allDescendants = this.matched
                .map((m) => m.descendants())
                .reduce((descendants, descendantPart) => {
                descendants.push(...descendantPart);
                return descendants;
            }, []);
            if (typeof queryInput === "string") {
                const query = (0, parser_1.tryQueryParser)(queryInput);
                if (query) {
                    return createQuery(this.document, query.match(allDescendants, this.document.descendants(), namespaces), this.virtualDoms, this);
                }
            }
            else {
                const elementMap = queryInput.matched.reduce((map, element) => {
                    map[element.identifier] = true;
                    return map;
                }, {});
                return createQuery(this.document, allDescendants.filter((m) => elementMap[m.identifier]), this.virtualDoms, this);
            }
            return this;
        };
        this.add = (toAdd, context, namespaces = {}) => {
            if (toAdd instanceof QueryInstance) {
                return createQuery(this.document, this.matched.concat(toAdd.matched), this.virtualDoms.concat(toAdd.virtualDoms), this);
            }
            else {
                const cssQuery = (0, parser_1.tryQueryParser)(toAdd);
                if (cssQuery) {
                    if (context) {
                        const nextMatched = [...this.matched];
                        context.matched.forEach((matched) => {
                            cssQuery
                                .match(matched.descendants(), this.document.descendants(), namespaces)
                                .forEach((element) => {
                                nextMatched.push(element);
                            });
                        });
                        return createQuery(this.document, nextMatched, this.virtualDoms, this);
                    }
                    else {
                        const nextMatched = [...this.matched];
                        cssQuery
                            .match(this.document.descendants(), this.document.descendants(), namespaces)
                            .forEach((element) => {
                            nextMatched.push(element);
                        });
                        return createQuery(this.document, nextMatched, this.virtualDoms, this);
                    }
                }
                else {
                    const htmlObject = (0, parser_1.tryHtmlParser)(toAdd, this.compress);
                    if (htmlObject) {
                        return createQuery(this.document, this.matched, this.virtualDoms.concat([htmlObject]), this);
                    }
                }
            }
            return this;
        };
        this.addBack = (selector, namespaces = {}) => {
            const query = selector && (0, parser_1.tryQueryParser)(selector);
            const matched = this.matched.concat(this.previous.matched);
            return createQuery(this.document, query
                ? query.match(matched, this.document.descendants(), namespaces)
                : matched, this.virtualDoms, this);
        };
        this.addClass = (classNames) => {
            this.matched.forEach((element, index) => {
                element.modifyAttribute("class", (classes) => {
                    const resolved = typeof classNames === "string"
                        ? classNames.split(" ")
                        : classNames(index, classes || "");
                    const arrayify = Array.isArray(resolved)
                        ? resolved
                        : resolved.split(" ");
                    if (!classes) {
                        return arrayify.join(" ");
                    }
                    const addedClasses = arrayify.filter((cls) => !classes.includes(cls));
                    const classList = classes.split(" ");
                    classList.push(...addedClasses);
                    return classList.join(" ");
                });
            });
            return this;
        };
        this.after = (firstArg, ...content) => {
            return this.addElementsToMatcher((content, matched, contentIndex) => {
                const tryFirstArgHtml = (0, parser_1.tryHtmlParser)(content, this.compress);
                if (!tryFirstArgHtml) {
                    if (matched.parent instanceof html_1.HtmlElement) {
                        matched.parent.content()?.addText(content, matched);
                    }
                }
                else {
                    const index = matched.parent.getIndex(matched, true);
                    tryFirstArgHtml.children().forEach((descentant) => {
                        matched.parent.addChild(descentant, index + contentIndex + 1);
                    });
                    const comments = (0, utils_1.getItems)(tryFirstArgHtml, (item) => item instanceof html_1.HtmlComment);
                    comments.forEach((comment) => {
                        if (comment.htmlComment?.value) {
                            matched.parent.addComment(comment.htmlComment.value, index + 1);
                        }
                        if (comment.htmlConditionalComment?.value) {
                            matched.parent.addConditionalComment(comment.htmlConditionalComment.value, index + 1);
                        }
                    });
                }
            }, firstArg, ...content);
        };
        this.append = (firstArg, ...content) => {
            return this.addElementsToMatcher((content, matched) => {
                const tryFirstArgHtml = (0, parser_1.tryHtmlParser)(content, this.compress);
                if (!tryFirstArgHtml) {
                    matched.convertWithChildren();
                    matched.content()?.addText(content);
                }
                else {
                    tryFirstArgHtml.children().forEach((descentant) => {
                        matched.convertWithChildren();
                        matched.addChild(descentant);
                    });
                    const comments = (0, utils_1.getItems)(tryFirstArgHtml, (item) => item instanceof html_1.HtmlComment);
                    comments.forEach((comment) => {
                        if (comment.htmlComment?.value) {
                            matched.parent.addComment(comment.htmlComment.value);
                        }
                        if (comment.htmlConditionalComment?.value) {
                            matched.parent.addConditionalComment(comment.htmlConditionalComment.value);
                        }
                    });
                }
            }, firstArg, ...content);
        };
        this.appendTo = (query, namespaces = {}) => {
            const parsedQuery = typeof query === "string"
                ? (0, parser_1.tryQueryParser)(query)?.match(this.document.descendants(), this.document.descendants(), namespaces)
                : query.matched;
            parsedQuery?.forEach((element) => {
                this.matched.forEach((nextChild) => {
                    element.convertWithChildren();
                    element.content()?.addChild(nextChild);
                    nextChild.parent.removeChild(nextChild);
                });
                this.virtualDoms.forEach((document) => {
                    document.children().forEach((child) => {
                        element.convertWithChildren();
                        element.content()?.addChild(child);
                    });
                    const comments = (0, utils_1.getItems)(document, (item) => item instanceof html_1.HtmlComment);
                    comments.forEach((comment) => {
                        if (comment.htmlComment?.value) {
                            element.convertWithChildren();
                            element.content()?.parent.addComment(comment.htmlComment.value);
                        }
                        if (comment.htmlConditionalComment?.value) {
                            element.convertWithChildren();
                            element
                                .content()
                                ?.parent.addConditionalComment(comment.htmlConditionalComment.value);
                        }
                    });
                });
            });
            this.document.cache.descendants.invalid = true;
            return this;
        };
        this.attr = (...args) => {
            const firstArg = args[0];
            if (typeof firstArg === "string") {
                if (args.length === 1 || args[1] === undefined) {
                    return this.matched[0]?.attributes()[firstArg] || "";
                }
                this.matched.forEach((matched, index) => {
                    const attributes = matched.attributes();
                    const resolveValue = typeof args[1] === "string"
                        ? args[1]
                        : args[1](index, attributes[firstArg] || "");
                    matched.replaceAttribute(firstArg, resolveValue);
                });
            }
            else if (typeof firstArg === "object" && firstArg) {
                this.matched.forEach((matched) => {
                    Object.entries(firstArg).forEach(([attributeName, attributeValue]) => {
                        matched.replaceAttribute(attributeName, attributeValue);
                    });
                });
            }
            return this;
        };
        this.before = (firstArg, ...content) => {
            return this.addElementsToMatcher((content, matched) => {
                const tryFirstArgHtml = (0, parser_1.tryHtmlParser)(content, this.compress);
                if (!tryFirstArgHtml) {
                    if (matched.parent instanceof html_1.HtmlContent) {
                        matched.parent.addText(content, matched);
                    }
                }
                else {
                    const index = matched.parent.getIndex(matched, true);
                    tryFirstArgHtml.children().forEach((descentant) => {
                        matched.parent.addChild(descentant, index);
                    });
                    const comments = (0, utils_1.getItems)(tryFirstArgHtml, (item) => item instanceof html_1.HtmlComment);
                    comments.forEach((comment) => {
                        if (comment.htmlComment?.value) {
                            matched.parent.addComment(comment.htmlComment.value, index);
                        }
                        if (comment.htmlConditionalComment?.value) {
                            matched.parent.addConditionalComment(comment.htmlConditionalComment.value, index);
                        }
                    });
                }
            }, firstArg, ...content);
        };
        this.blur = (handler) => this.on("blur", handler);
        this.change = (handler) => this.on("change", handler);
        this.children = () => {
            const children = (0, utils_1.flatten)(this.matched.map((m) => m.children()));
            return createQuery(this.document, children, this.virtualDoms, this);
        };
        this.clone = () => {
            const cloned = this.matched.map((m) => m.clone());
            const virtualDom = new html_1.HtmlDocument();
            cloned.forEach((clone) => virtualDom.addChild(clone));
            return createQuery(this.document, [], this.virtualDoms.concat([virtualDom]), this);
        };
        this.click = (handler) => this.on("click", handler);
        this.closest = (selector, container, namespaces = {}) => {
            const parsed = (0, parser_1.tryQueryParser)(selector);
            if (!parsed) {
                return this;
            }
            const descendantLimit = container?.matched.reduce((descendantsMap, matched) => {
                descendantsMap[matched.identifier] = true;
                matched.descendants().forEach((descendant) => {
                    descendantsMap[descendant.identifier] = true;
                });
                return descendantsMap;
            }, {});
            const seeker = (currentNode) => {
                if (descendantLimit && !descendantLimit[currentNode.identifier]) {
                    return undefined;
                }
                if (parsed.match([currentNode], this.document.descendants(), namespaces)
                    .length) {
                    return currentNode;
                }
                const parent = currentNode.parent;
                if (parent instanceof html_1.HtmlDocument) {
                    return undefined;
                }
                return seeker(parent);
            };
            const ancestors = this.matched.reduce((ancestors, element) => {
                const ancestor = seeker(element);
                if (ancestor) {
                    ancestors.push(ancestor);
                }
                return ancestors;
            }, []);
            return createQuery(this.document, ancestors, this.virtualDoms, this);
        };
        this.contents = () => {
            const children = this.matched.map((m) => (0, utils_1.parserItemToString)(m.tagClose?.close1?.closingGroup?.htmlContent) ||
                (0, utils_1.parserItemToString)(m.script) ||
                m.scriptlet?.value ||
                (0, utils_1.parserItemToString)(m.style));
            return children.join("");
        };
        this.contextmenu = (handler) => this.on("contextmenu", handler);
        this.css = (...args) => {
            const firstArg = args[0];
            if (typeof firstArg === "string") {
                if (args.length === 1 || args[1] === undefined) {
                    return this.matched[0]?.getStyles()[firstArg];
                }
                this.matched.forEach((matched, index) => {
                    const resolveValue = typeof args[1] === "string"
                        ? () => args[1]
                        : (value) => args[1](index, value);
                    matched.modifyStyle(firstArg, resolveValue);
                });
            }
            else if (Array.isArray(firstArg)) {
                const selectedMap = firstArg.reduce((selectedMap, item) => {
                    selectedMap[item] = true;
                    return selectedMap;
                }, {});
                return Object.entries(this.matched[0]?.getStyles() || {}).reduce((selected, [styleName, styleValue]) => {
                    if (selectedMap[styleName] && styleValue) {
                        selected[styleName] = styleValue;
                    }
                    return selected;
                }, {});
            }
            else if (typeof firstArg === "object" && firstArg) {
                this.matched.forEach((matched) => {
                    Object.entries(firstArg).forEach(([styleName, styleValue]) => {
                        matched.modifyStyle(styleName, () => styleValue);
                    });
                });
            }
            return this;
        };
        this.data = (...args) => {
            if (args.length === 0 || args[0] === undefined) {
                return this.matched[0]?.data || {};
            }
            else if (args.length === 1 || args[1] === undefined) {
                if (typeof args[0] === "string") {
                    return this.matched[0]?.data[args[0]];
                }
                if (typeof args[0] === "object" && args[0]) {
                    Object.keys(args[0]).forEach((key) => {
                        this.matched.forEach((matched) => {
                            matched.data[key] = args[0][key];
                        });
                    });
                }
            }
            else if (args.length === 2 || args[2] === undefined) {
                this.matched.forEach((matched) => {
                    matched.data[args[0]] = args[1];
                });
            }
            return this;
        };
        this.dblclick = (handler) => this.on("dblclick", handler);
        this.empty = () => {
            this.matched.forEach((matched) => {
                matched.emptyText();
                matched.emptyComments();
                matched.children().forEach((child) => {
                    matched.content()?.removeChild(child);
                });
            });
            return this;
        };
        this.end = () => this.previous;
        this.eq = (index) => {
            return createQuery(this.document, this.matched.filter((_, i) => index >= 0 ? i === index : i === this.matched.length - 1 + index), this.virtualDoms, this);
        };
        this.filter = (filter, namespaces = {}) => {
            if (typeof filter === "string") {
                const query = (0, parser_1.tryQueryParser)(filter);
                if (query) {
                    return createQuery(this.document, query.match(this.matched, this.document.descendants(), namespaces), this.virtualDoms, this);
                }
            }
            else if (filter instanceof QueryInstance) {
                const matchedMap = filter.matched.reduce((matchedMap, element) => {
                    matchedMap[element.identifier] = true;
                    return matchedMap;
                }, {});
                return createQuery(this.document, this.matched.filter((m) => matchedMap[m.identifier]), this.virtualDoms, this);
            }
            else if (typeof filter === "function") {
                return createQuery(this.document, this.matched.filter((m, i) => filter(i, createQuery(this.document, [m], this.virtualDoms, this))), this.virtualDoms, this);
            }
            return this;
        };
        this.each = (callback) => {
            for (let i = 0; i < this.matched.length; i++) {
                callback(i, this.eq(i));
            }
            return this;
        };
        this.even = () => {
            return createQuery(this.document, this.matched.filter((_, i) => i % 2 === 0), this.virtualDoms, this);
        };
        this.first = () => this.eq(0);
        this.focus = (handler) => this.on("focus", handler);
        this.focusin = (handler) => this.on("focusin", handler);
        this.focusout = (handler) => this.on("focusout", handler);
        this.has = (selector, namespaces = {}) => {
            if (typeof selector === "string") {
                const query = (0, parser_1.tryQueryParser)(selector);
                if (query) {
                    return createQuery(this.document, this.matched.filter((m) => {
                        return (query.match(m.descendants(), this.document.descendants(), namespaces).length > 0);
                    }), this.virtualDoms, this);
                }
            }
            else if (selector instanceof QueryInstance) {
                const matchedMap = selector.matched.reduce((matchedMap, element) => {
                    matchedMap[element.identifier] = true;
                    return matchedMap;
                }, {});
                return createQuery(this.document, this.matched.filter((m) => {
                    return m.descendants().some((d) => matchedMap[d.identifier]);
                }), this.virtualDoms, this);
            }
            return this;
        };
        this.hasClass = (className) => {
            return this.matched.some((m) => m
                .attributes()["class"]?.split(" ")
                .some((cls) => cls === className));
        };
        this.hide = () => {
            this.matched.forEach((m) => {
                m.modifyStyle("display", () => "none");
            });
            return this;
        };
        this.hover = (handler) => this.on("hover", handler);
        this.html = (...args) => {
            if (args.length === 0 || args[0] === undefined) {
                if (this.matched.length > 0) {
                    return (0, utils_1.parserItemToString)(this.matched[0].content());
                }
                else {
                    return "";
                }
            }
            else {
                this.matched.forEach((m, index) => {
                    const resolveArg = typeof args[0] === "string"
                        ? args[0]
                        : args[0](index, (0, utils_1.parserItemToString)(m.content()));
                    const nextDom = (0, parser_1.tryHtmlParser)(resolveArg, this.compress);
                    if (nextDom) {
                        m.convertWithChildren();
                        if (m.tagClose?.close1?.closingGroup) {
                            m.tagClose.close1.closingGroup.htmlContent = new html_1.HtmlContent(m);
                        }
                        nextDom.children().forEach((d) => {
                            m.addChild(d);
                        });
                    }
                });
                this.document.cache.children.invalid = true;
                this.document.cache.descendants.invalid = true;
                this.document.cache.indexes.invalid = true;
            }
            return this;
        };
        this.index = (selector, namespaces = {}) => {
            if (selector === undefined) {
                return this.matched[0]
                    ? this.matched[0].parent.getIndex(this.matched[0], false)
                    : -1;
            }
            if (typeof selector === "string") {
                const query = (0, parser_1.tryQueryParser)(selector);
                if (query) {
                    const matched = query.match(this.matched, this.document.descendants(), namespaces);
                    if (matched.length > 0) {
                        return this.matched.findIndex((m) => m.identifier === matched[0].identifier);
                    }
                }
            }
            if (selector instanceof QueryInstance) {
                const firstMatch = selector.matched[0];
                if (firstMatch) {
                    return this.matched.findIndex((m) => m.identifier === firstMatch.identifier);
                }
            }
            return -1;
        };
        this.insertBefore = (selector, namespaces = {}) => {
            const toInsertBefore = typeof selector === "string"
                ? (() => {
                    const query = (0, parser_1.tryQueryParser)(selector);
                    if (query) {
                        return query.match(this.document.descendants(), this.document.descendants(), namespaces);
                    }
                    return [];
                })()
                : selector.matched;
            toInsertBefore.forEach((m) => {
                this.matched.forEach((matched, matchedIndex) => {
                    m.parent.addChild(matched, m.parent.getIndex(m, true) + matchedIndex);
                });
                this.virtualDoms.forEach((v) => {
                    v.children().forEach((virtual, virtualIndex) => {
                        m.parent.addChild(virtual, Math.max(0, m.parent.getIndex(m, true) +
                            this.matched.length -
                            1 +
                            virtualIndex));
                    });
                });
            });
            this.document.cache.descendants.invalid = true;
            return createQuery(this.document, this.matched, this.virtualDoms, this);
        };
        this.insertAfter = (selector, namespaces = {}) => {
            const toInsertAfter = typeof selector === "string"
                ? (() => {
                    const query = (0, parser_1.tryQueryParser)(selector);
                    if (query) {
                        return query.match(this.document.descendants(), this.document.descendants(), namespaces);
                    }
                    return [];
                })()
                : selector.matched;
            toInsertAfter.forEach((m) => {
                const mIndex = m.parent.getIndex(m, true);
                this.matched.forEach((matched, matchedIndex) => {
                    m.parent.addChild(matched, mIndex + matchedIndex + 1);
                });
                this.virtualDoms.forEach((v) => {
                    v.children().forEach((virtual, virtualIndex) => {
                        m.parent.addChild(virtual, mIndex + this.matched.length + virtualIndex + 1);
                    });
                });
            });
            this.document.cache.descendants.invalid = true;
            return createQuery(this.document, this.matched, [], this);
        };
        this.keypress = (handler) => this.on("keypress", handler);
        this.keydown = (handler) => this.on("keydown", handler);
        this.keyup = (handler) => this.on("keyup", handler);
        this.is = (selector, namespaces = {}) => {
            if (typeof selector === "string") {
                const query = (0, parser_1.tryQueryParser)(selector);
                if (query) {
                    return (query.match(this.matched, this.document.descendants(), namespaces)
                        .length > 0);
                }
                return false;
            }
            const map = selector.matched.reduce((map, element) => {
                map[element.identifier] = true;
                return map;
            }, {});
            return this.matched.some((m) => map[m.identifier]);
        };
        this.last = () => this.eq(this.matched.length - 1);
        this.map = (iterator) => {
            return this.matched.map((m, i) => iterator(i, createQuery(this.document, [m], this.virtualDoms, this)));
        };
        this.mousedown = (handler) => this.on("mousedown", handler);
        this.mouseenter = (handler) => this.on("mouseenter", handler);
        this.mouseleave = (handler) => this.on("mouseleave", handler);
        this.mousemove = (handler) => this.on("mousemove", handler);
        this.mouseout = (handler) => this.on("mouseout", handler);
        this.mouseover = (handler) => this.on("mouseover", handler);
        this.mouseup = (handler) => this.on("mouseup", handler);
        this.next = (selector, namespaces = {}) => {
            const nextSiblings = this.matched
                .map((m) => m.parent.nextSibling(m))
                .filter(Boolean);
            return this.filterBySelector(selector, nextSiblings, namespaces);
        };
        this.nextAll = (selector, namespaces = {}) => {
            const nextSiblings = (0, utils_1.flatten)(this.matched.map((m) => {
                const index = m.parent.getIndex(m, false);
                return m.parent
                    .children()
                    .filter((child) => child.parent.getIndex(child, false) > index);
            }));
            return this.filterBySelector(selector, nextSiblings, namespaces);
        };
        this.nextUntil = (selector, filter, namespaces = {}) => {
            const until = this.getUntil(selector, namespaces);
            const nextSiblings = (0, utils_1.flatten)(this.matched.map((m) => {
                const index = m.parent.getIndex(m, false);
                return m.parent
                    .children()
                    .filter((child) => child.parent.getIndex(child, false) > index)
                    .reduce(({ hasMatchedSelector, selected }, element) => {
                    if (hasMatchedSelector || until[element.identifier]) {
                        return { hasMatchedSelector: true, selected };
                    }
                    selected.push(element);
                    return { hasMatchedSelector, selected };
                }, { hasMatchedSelector: false, selected: new Array() }).selected;
            }));
            return this.filterBySelector(filter, nextSiblings, namespaces);
        };
        this.not = (selector, namespaces = {}) => {
            if (typeof selector === "function") {
                return this.filter((index, instance) => !selector(index, instance));
            }
            const filtered = this.filterBySelector(selector, this.matched, namespaces).matched.reduce((map, element) => {
                map[element.identifier] = true;
                return map;
            }, {});
            return createQuery(this.document, this.matched.filter((m) => !filtered[m.identifier]), this.virtualDoms, this);
        };
        this.odd = () => {
            return createQuery(this.document, this.matched.filter((_, i) => i % 2 === 1), this.virtualDoms, this);
        };
        this.off = (eventTypes) => {
            this.matched.forEach((m) => {
                eventTypes.split(" ").forEach((eventType) => {
                    m.removeAttribute(`on${eventType}`);
                });
            });
            return this;
        };
        this.on = (eventTypes, eventHandler) => {
            const stringifyHandler = typeof eventHandler === "function"
                ? `(${eventHandler.toString()})(this)`
                : eventHandler;
            this.matched.forEach((match) => {
                eventTypes.split(" ").forEach((eventType) => {
                    match.modifyAttribute(`on${eventType}`, (existing) => {
                        if (!existing) {
                            return stringifyHandler;
                        }
                        const split = existing.split(";");
                        split.push(stringifyHandler);
                        return split.join(";");
                    });
                });
            });
            return this;
        };
        this.once = (eventTypes, eventHandler) => {
            eventTypes.split(" ").forEach((eventType) => {
                const stringifyHandler = typeof eventHandler === "function"
                    ? eventHandler.toString()
                    : eventHandler;
                const identifier = (0, utils_1.uniqueId)("once_event_handler");
                const onceHandler = `((event) => { if (window['${identifier}']) { return; } const fn = ${stringifyHandler}; const result = fn(event); window['${identifier}'] = true; return result; })(this)`;
                this.on(eventType, onceHandler);
            });
            return this;
        };
        this.parent = (selector, namespaces = {}) => {
            return this.filterBySelector(selector, this.matched.reduce((parents, element) => {
                const parentContent = element.parent;
                if (parentContent instanceof html_1.HtmlElement) {
                    parents.push(parentContent);
                }
                return parents;
            }, []), namespaces);
        };
        this.parents = (selector, namespaces = {}) => {
            return this.filterBySelector(selector, this.matched.reduce((parents, element) => {
                let parentContent = element.parent;
                while (parentContent instanceof html_1.HtmlElement) {
                    parents.push(parentContent);
                    parentContent = parentContent.parent;
                }
                return parents;
            }, []), namespaces);
        };
        this.parentsUntil = (selector, namespaces = {}) => {
            const until = this.getUntil(selector, namespaces);
            const parents = this.matched.reduce((parents, element) => {
                let parentContent = element.parent;
                while (parentContent instanceof html_1.HtmlElement &&
                    !until[parentContent.identifier]) {
                    parents.push(parentContent);
                    parentContent = parentContent.parent;
                }
                return parents;
            }, []);
            return createQuery(this.document, parents, this.virtualDoms, this);
        };
        this.prepend = (firstArg, ...content) => {
            return this.addElementsToMatcher((content, matched, prependIndex) => {
                const tryFirstArgHtml = (0, parser_1.tryHtmlParser)(content, this.compress);
                if (!tryFirstArgHtml) {
                    matched.convertWithChildren();
                    matched.content()?.addText(content);
                }
                else {
                    tryFirstArgHtml.children().forEach((child, childIndex) => {
                        matched.convertWithChildren();
                        matched.content()?.addChild(child, childIndex + prependIndex);
                    });
                }
            }, firstArg, ...content);
        };
        this.prependTo = (query, namespaces = {}) => {
            const matched = typeof query === "string"
                ? (0, parser_1.tryQueryParser)(query)?.match(this.document.descendants(), this.document.descendants(), namespaces)
                : query.matched;
            matched?.forEach((element) => {
                this.matched.forEach((nextChild) => {
                    element.convertWithChildren();
                    element.addChild(nextChild, 0);
                    nextChild.parent.removeChild(nextChild);
                });
                this.virtualDoms.forEach((document) => {
                    document.children().forEach((child) => {
                        element.convertWithChildren();
                        element.addChild(child, 0);
                    });
                });
            });
            this.document.cache.descendants.invalid = true;
            return createQuery(this.document, this.matched, [], this);
        };
        this.prev = (selector, namespaces = {}) => {
            const prevSiblings = this.matched
                .map((m) => m.parent.prevSibling(m))
                .filter(Boolean);
            return this.filterBySelector(selector, prevSiblings, namespaces);
        };
        this.prevAll = (selector, namespaces = {}) => {
            const prevSiblings = (0, utils_1.flatten)(this.matched.map((m) => {
                const index = m.parent.getIndex(m, false);
                return m.parent
                    .children()
                    .filter((child) => child.parent.getIndex(child, false) < index);
            }));
            return this.filterBySelector(selector, prevSiblings, namespaces);
        };
        this.prevUntil = (selector, filter, namespaces = {}) => {
            const until = this.getUntil(selector, namespaces);
            const prevSiblings = (0, utils_1.flatten)(this.matched.map((m) => {
                const index = m.parent.getIndex(m, false);
                return m.parent
                    .children()
                    .filter((child) => child.parent.getIndex(child, false) < index)
                    .reverse()
                    .reduce(({ hasMatchedSelector, selected }, element) => {
                    if (hasMatchedSelector || until[element.identifier]) {
                        return { hasMatchedSelector: true, selected };
                    }
                    selected.push(element);
                    return { hasMatchedSelector, selected };
                }, { hasMatchedSelector: false, selected: new Array() }).selected;
            }));
            return this.filterBySelector(filter, prevSiblings, namespaces);
        };
        this.print = (ignoreWhitespace) => {
            const printed = this.matched
                .map(utils_1.parserItemToString)
                .concat(this.virtualDoms.map(utils_1.parserItemToString))
                .join("");
            return ignoreWhitespace
                ? printed.replace(/|\n\s*/gu, "").replace(/\s+/gu, " ")
                : printed;
        };
        /**
         * Using this instead of prop()
         */
        this.tagName = (...args) => {
            if (args.length === 0 || args[0] === undefined) {
                return this.matched[0]?.tagName?.value;
            }
            this.matched.forEach((m, i) => {
                const resolveArg = typeof args[0] === "string" ? args[0] : args[0](i, m.tagName?.value || "");
                m.tagName = new types_1.LexerItem("TAG_NAME", resolveArg);
                if (m.tagClose?.close1?.closingGroup) {
                    m.tagClose.close1.closingGroup.tagName = new types_1.LexerItem("TAG_NAME", resolveArg);
                }
            });
            return this;
        };
        this.pushStack = (instance) => createQuery(this.document, this.matched.concat(instance.matched), this.virtualDoms, this);
        this.ready = (handler) => {
            if (this.matched.length) {
                return this.on("load", handler);
            }
            else {
                const resolveHandler = typeof handler === "string" ? handler : handler.toString();
                const wrapHandler = `document.addEventListener("load", ${resolveHandler})`;
                const script = (0, parser_1.htmlParser)(`<script>${wrapHandler}</script>`, this.compress).descendants()[0];
                const query = (0, parser_1.queryParser)("body");
                const body = query.match(this.document.descendants(), this.document.descendants(), {})[0];
                if (body) {
                    body.convertWithChildren();
                    body.addChild(script);
                }
            }
            this.document.cache.descendants.invalid = true;
            return this;
        };
        this.reduce = (reducer, initialValue) => {
            for (let i = 0; i < this.matched.length; i++) {
                initialValue = reducer(initialValue, this.eq(i), i, this);
            }
            return initialValue;
        };
        this.remove = (selector, namespaces = {}) => {
            if (selector) {
                const query = (0, parser_1.tryQueryParser)(selector);
                if (query) {
                    const toDelete = query.match(this.matched, this.document.descendants(), namespaces);
                    toDelete.forEach((element) => {
                        element.parent.removeChild(element);
                    });
                    const deleteMap = toDelete.reduce((map, element) => {
                        map[element.identifier] = true;
                        return map;
                    }, {});
                    const nextMatched = this.matched.filter((m) => !deleteMap[m.identifier]);
                    this.document.cache.descendants.invalid = true;
                    return createQuery(this.document, nextMatched, this.virtualDoms, this);
                }
            }
            else {
                this.matched.forEach((m) => m.parent.removeChild(m));
                this.document.cache.descendants.invalid = true;
                return createQuery(this.document, [], this.virtualDoms, this);
            }
            return this;
        };
        this.removeAttr = (attributeNames) => {
            this.matched.forEach((m) => attributeNames
                .split(" ")
                .forEach((attributeName) => m.removeAttribute(attributeName)));
            return this;
        };
        this.removeClass = (classNames) => {
            this.matched.forEach((m, index) => {
                m.modifyAttribute("class", (attribute) => {
                    if (!attribute) {
                        return attribute;
                    }
                    const resolvedClassNames = typeof classNames === "string"
                        ? classNames.split(" ")
                        : classNames(index, attribute);
                    const bannedClassNames = (typeof resolvedClassNames === "string"
                        ? resolvedClassNames.split(" ")
                        : resolvedClassNames).reduce((banned, className) => {
                        banned[className] = true;
                        return banned;
                    }, {});
                    return attribute
                        .split(" ")
                        .filter((className) => !bannedClassNames[className])
                        .join(" ");
                });
            });
            return this;
        };
        this.removeData = (data) => {
            const arrayified = typeof data === "string" ? [data] : data;
            arrayified.forEach((value) => {
                this.matched.forEach((m) => {
                    delete m.data[value];
                });
            });
            return this;
        };
        this.replaceAll = (replacement, namespaces = {}) => {
            const arrayified = Array.isArray(replacement) ? replacement : [replacement];
            const elements = (0, utils_1.flatten)(arrayified.map((item) => {
                if (typeof item === "string") {
                    const query = (0, parser_1.tryQueryParser)(item);
                    return (query?.match(this.document.descendants(), this.document.descendants(), namespaces) || []);
                }
                else {
                    return item.matched;
                }
            }));
            const replacements = this.matched.concat((0, utils_1.flatten)(this.virtualDoms.map((v) => v.children())));
            elements.forEach((element) => {
                const parent = element.parent;
                const index = parent.getIndex(element, true);
                parent.removeChild(element);
                replacements.forEach((replacement) => {
                    parent.addChild(replacement.clone(), index);
                });
            });
            this.document.cache.descendants.invalid = true;
            return this;
        };
        this.replaceWith = (content) => {
            const executed = typeof content === "function" ? content() : content;
            const arrayified = Array.isArray(executed) ? executed : [executed];
            const elements = (0, utils_1.flatten)(arrayified.map((item) => {
                if (typeof item === "string") {
                    return (0, parser_1.tryHtmlParser)(item, this.compress)?.children() || [];
                }
                else {
                    return item.matched;
                }
            }));
            this.matched.forEach((m) => {
                const parent = m.parent;
                const index = parent.getIndex(m, true);
                elements.forEach((e) => {
                    parent.addChild(e.clone(), index);
                });
            });
            this.document.cache.descendants.invalid = true;
            return createQuery(this.document, [], this.virtualDoms, this);
        };
        this.select = (handler) => this.on("select", handler);
        this.siblings = (selector, namespaces = {}) => {
            const matched = (0, utils_1.flatten)(this.matched.map((m) => {
                const index = m.parent.getIndex(m, false);
                return m.parent.children().filter((_, i) => i !== index);
            }));
            return this.filterBySelector(selector, matched, namespaces);
        };
        this.slice = (from, to) => {
            const matched = this.matched.slice(from, to);
            return new QueryInstance(this.document, matched, this.virtualDoms, this);
        };
        this.submit = (handler) => this.on("submit", handler);
        this.text = (...args) => {
            if (args.length === 0 || args[0] === undefined) {
                return this.matched[0]?.texts().join(" ");
            }
            if (args.length === 1 || args[1] === undefined) {
                this.matched.forEach((m, index) => {
                    const value = (typeof args[0] === "function"
                        ? args[0](index, m.texts().join(" "))
                        : args[0]).toString();
                    m.convertWithChildren();
                    m.emptyText();
                    m.addText(value);
                });
            }
            return this;
        };
        this.toggleClass = (classNames, state) => {
            if (state === true) {
                return this.addClass(classNames);
            }
            if (state === false) {
                return this.removeClass(classNames);
            }
            this.matched.forEach((m, index) => {
                m.modifyAttribute("class", (classes) => {
                    const resolve = typeof classNames === "function"
                        ? classNames(index, classes || "")
                        : classNames.split(" ");
                    const arrayify = Array.isArray(resolve)
                        ? (0, utils_1.flatten)(resolve.map((r) => r.split(" ")))
                        : resolve.split(" ");
                    const classList = (classes || "").split(" ");
                    const map = classList.reduce((map, cls) => {
                        map[cls] = true;
                        return map;
                    }, {});
                    const found = arrayify.filter((cls) => map[cls]);
                    const notFound = arrayify.filter((cls) => !map[cls]);
                    const toRemove = found.reduce((toRemove, cls) => {
                        toRemove[cls] = true;
                        return toRemove;
                    }, {});
                    const filtered = classList.filter((cls) => !toRemove[cls]);
                    filtered.push(...notFound);
                    return filtered.join(" ");
                });
            });
            return this;
        };
        this.uniqueSort = () => {
            const unique = Object.values(this.matched.reduce((map, element) => {
                map[element.identifier] = element;
                return map;
            }, {}));
            const indexed = unique.map((element) => {
                const indexes = [];
                const seeker = (element) => {
                    const index = element.parent.getIndex(element, false);
                    if (element.parent instanceof html_1.HtmlDocument) {
                        indexes.push(index);
                    }
                    else {
                        indexes.push(index);
                        seeker(element.parent);
                    }
                };
                seeker(element);
                return {
                    element,
                    indexes,
                };
            });
            const arrayComparator = (aArray, bArray) => {
                const minLength = Math.min(aArray.length, bArray.length);
                const aCommon = aArray.slice(0, minLength);
                const bCommon = bArray.slice(0, minLength);
                const diffIndex = aCommon.findIndex((a, i) => a !== bCommon[i]);
                if (diffIndex === -1) {
                    return aArray.length - bArray.length;
                }
                return aCommon[diffIndex] - bCommon[diffIndex];
            };
            const sorted = indexed.sort(({ indexes: aIndex }, { indexes: bIndex }) => {
                return arrayComparator(aIndex, bIndex);
            });
            return createQuery(this.document, sorted.map(({ element }) => element), this.virtualDoms, this);
        };
        this.unload = (handler) => this.on("unload", handler);
        this.unwrap = (selector, namespaces = {}) => {
            const toUnwrap = this.matched.reduce((toUnwrap, element) => {
                if (element.parent instanceof html_1.HtmlElement) {
                    toUnwrap.push({ element, parent: element.parent });
                }
                return toUnwrap;
            }, []);
            const elements = (selector === undefined
                ? toUnwrap.map(({ parent }) => parent)
                : typeof selector === "string"
                    ? (() => {
                        const parents = toUnwrap.map((u) => u.parent);
                        const query = (0, parser_1.tryQueryParser)(selector);
                        if (!query) {
                            return parents;
                        }
                        return query.match(parents, this.document.descendants(), namespaces);
                    })()
                    : selector.matched).reduce((elements, element) => {
                elements[element.identifier] = true;
                return elements;
            }, {});
            const unwrap = (element) => {
                if (element.parent instanceof html_1.HtmlDocument) {
                    return;
                }
                const grandparent = element.parent.parent;
                const siblings = element.parent.children();
                const parentIndex = grandparent.getIndex(element.parent, true);
                grandparent.removeChild(parentIndex);
                siblings.forEach((sibling) => grandparent.addChild(sibling));
            };
            toUnwrap
                .filter(({ parent }) => elements[parent.identifier])
                .forEach(({ element }) => {
                unwrap(element);
            });
            this.document.cache.descendants.invalid = true;
            return this;
        };
        this.val = (...args) => {
            if (args.length === 0 || args[0] === undefined) {
                return this.matched[0]?.attributes()["value"];
            }
            this.attr("value");
            return this.attr("value", args[0]);
        };
        this.wrap = (...args) => {
            this.matched.forEach((m, index) => {
                const resolved = typeof args[0] === "function" ? args[0](index, m) : args[0];
                const namespaces = args[1] || {};
                const [parent, appendTo] = this.getParentAssigner(resolved, namespaces);
                if (parent && appendTo) {
                    appendTo.convertWithChildren();
                    appendTo.addChild(m.clone());
                    m.parent.replaceChild(m, parent);
                }
            });
            this.document.cache.descendants.invalid = true;
            return this;
        };
        this.wrapAll = (...args) => {
            if (!this.matched.length) {
                return this;
            }
            const getParentList = (element, parents = {}, index = 0) => {
                parents[element.identifier] = { element, index };
                return element instanceof html_1.HtmlDocument
                    ? parents
                    : getParentList(element.parent, parents, index + 1);
            };
            const parentLists = this.matched.map((m) => getParentList(m));
            const allParentKeys = Object.keys((0, utils_1.flatten)(parentLists.map((p) => Object.keys(p))).reduce((acc, toUnique) => {
                acc[toUnique] = true;
                return acc;
            }, {}));
            const onlyCommonKeys = allParentKeys.filter((key) => parentLists.every((parent) => parent[key] && parent[key].index !== 0));
            const commonKeysWithLowestIndex = onlyCommonKeys.map((key) => {
                return {
                    key,
                    index: Math.min(...parentLists.map((p) => p[key].index)),
                };
            });
            const lowestCommonKey = commonKeysWithLowestIndex.sort(({ index: aIndex }, { index: bIndex }) => {
                return aIndex - bIndex;
            })[0];
            const commonParent = parentLists[0][lowestCommonKey.key].element;
            const childrenToBeWrapped = commonParent.children().filter((element) => {
                return parentLists.some((list) => list[element.identifier]);
            });
            const childrenToBeWrappedMap = childrenToBeWrapped.reduce((childrenToBeWrappedMap, child) => {
                childrenToBeWrappedMap[child.identifier] = true;
                return childrenToBeWrappedMap;
            }, {});
            const resolved = typeof args[0] === "function" ? args[0]() : args[0];
            const element = typeof resolved === "string"
                ? (0, parser_1.tryHtmlParser)(resolved, this.compress)?.children()[0]
                : resolved.matched[0];
            if (element) {
                if (commonParent instanceof html_1.HtmlDocument) {
                    const copy = [...commonParent.htmlElements]
                        .filter((element) => element.htmlElement?.consumed() &&
                        childrenToBeWrappedMap[element.htmlElement.identifier])
                        .map((element) => element.clone());
                    copy.forEach((c) => {
                        const convertMisc = (misc) => {
                            if (misc.htmlComment?.htmlComment?.value) {
                                element.convertWithChildren();
                                element.addComment(misc.htmlComment.htmlComment.value);
                            }
                            if (misc.htmlComment?.htmlConditionalComment?.value) {
                                element.convertWithChildren();
                                element.addConditionalComment(misc.htmlComment.htmlConditionalComment.value);
                            }
                            if (misc.seaWs?.value) {
                                element.convertWithChildren();
                                element.content()?.addText(misc.seaWs.value);
                            }
                        };
                        c.htmlMisc1.forEach(convertMisc);
                        if (c.htmlElement) {
                            element.addChild(c.htmlElement);
                        }
                        c.htmlMisc2.forEach(convertMisc);
                    });
                    const wrapper = new html_1.HtmlElements(commonParent);
                    wrapper.htmlElement = element;
                    const rest = [
                        wrapper,
                        ...commonParent.htmlElements.filter((element) => element.htmlElement && !childrenToBeWrappedMap[element.htmlElement.identifier]),
                    ];
                    commonParent.htmlElements = rest;
                }
                else {
                    commonParent.convertWithChildren();
                    element.convertWithChildren();
                    const copy = commonParent
                        .children()
                        .filter((c) => childrenToBeWrappedMap[c.identifier])
                        .map((c) => c.clone());
                    copy.forEach((c) => element.addChild(c));
                    const rest = commonParent
                        .content()
                        ?.content.filter((c) => c instanceof html_1.HtmlElement && !childrenToBeWrappedMap[c.identifier]);
                    if (commonParent.tagClose?.close1?.closingGroup) {
                        commonParent.tagClose.close1.closingGroup.htmlContent = new html_1.HtmlContent(commonParent);
                        commonParent.addChild(element);
                        commonParent.content()?.content.push(...(rest || []));
                    }
                }
            }
            this.document.cache.descendants.invalid = true;
            return this;
        };
        this.wrapInner = (...args) => {
            this.matched.forEach((m, index) => {
                const resolved = typeof args[0] === "function" ? args[0](index, m) : args[0];
                const namespaces = args[1] || {};
                const [parent, appendTo] = this.getParentAssigner(resolved, namespaces);
                if (parent && appendTo) {
                    const content = m.content()?.clone();
                    m.convertWithChildren();
                    if (m.tagClose?.close1?.closingGroup) {
                        m.tagClose.close1.closingGroup.htmlContent = new html_1.HtmlContent(m);
                    }
                    appendTo.convertWithChildren();
                    if (appendTo.tagClose?.close1?.closingGroup) {
                        appendTo.tagClose.close1.closingGroup.htmlContent = content;
                    }
                    m.addChild(parent);
                }
            });
            this.document.cache.descendants.invalid = true;
            return this;
        };
        this.document = document;
        this.matched = Object.values(matched.reduce((matched, element) => {
            matched[element.identifier] = element;
            return matched;
        }, {}));
        this.virtualDoms = virtualDoms;
        if (previous instanceof QueryInstance) {
            this.previous = previous;
            this.compress = previous.compress;
        }
        else {
            this.previous = this;
            this.compress = previous.compress;
        }
    }
    *[Symbol.iterator]() {
        for (let i = 0; i < this.matched.length; i++) {
            yield this.eq(i);
        }
    }
    get length() {
        return this.matched.length;
    }
}
const createQuery = (document, matched, virtualDoms, previous) => {
    const instance = new QueryInstance(document, matched, virtualDoms, previous);
    return new Proxy(instance, {
        get(target, propertyKey, receiver) {
            const tryParseKey = parseInt(propertyKey.toString());
            if (!isNaN(tryParseKey)) {
                return target.eq(tryParseKey);
            }
            return Reflect.get(target, propertyKey, receiver);
        },
        apply(target, _, args) {
            return target.ready(args[0]);
        },
    });
};
const Query = (htmlInput, compress = true) => {
    const html = (0, parser_1.htmlParser)(htmlInput, compress);
    return (queryInput, namespaces = {}) => {
        if (typeof queryInput === "function") {
            const query = createQuery(html, [], [], { compress });
            return query.ready(queryInput);
        }
        if (!queryInput) {
            return createQuery(html, [], [], { compress });
        }
        const query = (0, parser_1.tryQueryParser)(queryInput);
        if (query) {
            const allHtmLElements = html.descendants();
            const matched = query.match(allHtmLElements, allHtmLElements, namespaces);
            return createQuery(html, matched, [], { compress });
        }
        else {
            const virtualDom = (0, parser_1.tryHtmlParser)(queryInput, compress);
            if (virtualDom) {
                return createQuery(html, [], [virtualDom], { compress });
            }
            return createQuery(html, [], [], { compress }).ready(queryInput);
        }
    };
};
exports.Query = Query;
